import os
import shutil

import pytest
from src.tmux_conf import TmuxConfig
from src.tmux_conf.exceptions import TmuxConfNotTmuxCommand
from src.tmux_conf.utils import run_shell

from .utils_test import CONF_FILE, install_plugins, remove_conf_file, tmux_conf_instance

CONTENT_LINE = "# Hello world"
PLUGIN_SOURCE = "jaclu"
PLUGIN_NAME = "tmux-prefix-highlight"


class WithContent(TmuxConfig):
    tst_content = "# dummy content"

    def content(self):
        self.write(self.tst_content)


class NotEmbedded(WithContent):
    """We combine not embedded and content in the same class.
    They are used for separate tests so they dont overlap,
    just handy to use a single class"""

    use_embedded_scripts = False


class PluginsSample(WithContent):
    """Test class ensuring plugins are only present at the intended
    versions"""

    def plugin_prefix_highlight(self):
        #
        #  Highlights when you press tmux prefix key and
        #  when copy/sync mode is active.
        #
        conf = """
        set -g @prefix_highlight_show_copy_mode  on
        set -g @prefix_highlight_copy_mode_attr  "fg=black,bg=yellow,bold"
        set -g @prefix_highlight_show_sync_mode  on
        set -g @prefix_highlight_sync_mode_attr "fg=black,bg=orange,blink,bold"
        """
        return [f"{PLUGIN_SOURCE}/{PLUGIN_NAME}", 2.4, conf]

    def plugin_session_wizard(self):  # 3.2
        return ["27medkamal/tmux-session-wizard", 3.2, "#  Default trigger: <prefix> T"]


class PluginsDisabled(PluginsSample):
    """Test class ensuring no plugins show up if plugin_handler is
    disabled"""

    plugin_handler = ""


class PluginsManual(PluginsSample):
    """Test class ensuring no plugins show up if plugin_handler is
    disabled"""

    plugin_handler = "manual"


class PluginsDuplicated(PluginsSample):
    def plugin_prefix_highlight2(self):
        #
        #  Highlights when you press tmux prefix key and
        #  when copy/sync mode is active.
        #
        conf = """
        set -g @prefix_highlight_show_copy_mode  on
        set -g @prefix_highlight_copy_mode_attr  "fg=black,bg=yellow,bold"
        set -g @prefix_highlight_show_sync_mode  on
        set -g @prefix_highlight_sync_mode_attr "fg=black,bg=orange,blink,bold"
        """
        return [f"{PLUGIN_SOURCE}/{PLUGIN_NAME}", 2.4, conf]


def tc_env(conf_file=CONF_FILE):
    tc = TmuxConfig(parse_cmd_line=False, conf_file=conf_file)
    return tc


def prep_plugin_class(cls, version="2.8", plugins_display=0):
    t = tmux_conf_instance(
        cls=cls, tmux_version=version, plugins_display=plugins_display
    )
    # ps = cls(parse_cmd_line=False, conf_file=CONF_FILE, tmux_version=version)
    t.plugins.scan(t.list_plugin_methods())
    return t


def not_test_tc_vers_ok_low():
    tc = tc_env()
    assert tc.vers_ok(1.0) is True


def not_test_tc_vers_ok_high():
    tc = tc_env()
    assert tc.vers_ok(99) is not True


def not_test_tc_find_tmux_bin():
    tc = tc_env()
    assert tc.find_tmux_bin() is True


# def not_test_tc_is_tmux_bin():
#     tc = tc_env()
#     #  This assumes there is a tmux in PATH
#     assert tc.is_tmux_bin("tmux") is True


#
#  Cant be run inside VSCode ATM...
#
# cfb = os.environ.get("__CFBundleIdentifier")
# if not cfb or cfb.find("VSCode") < 0:

#     def not_test_tc_parse_cmd_line():
#         remove_conf_file()
#         t = TmuxConfig()
#         with pytest.raises(OSError):
#             t.run()


def not_test_tc_content_needs_overloading():
    t = TmuxConfig(parse_cmd_line=False)
    with pytest.raises(NotImplementedError) as exc:
        t.content()
    assert exc.value.args[0] == "content() must be defined in sub-class!"


def not_test_tc_repeated_calls_to_edit_config_coverage():
    t = TmuxConfig(parse_cmd_line=False)
    t.edit_config()
    t.edit_config()


def not_test_tc_btick():
    t = tmux_conf_instance(cls=WithContent)
    t.tst_content = "# one btick `"
    with pytest.raises(SyntaxError) as exc:
        t.run()
    assert exc.value.args[0].find("Un-escaped back-ticks can not be present") > -1


#
#  Handle notes
#
# def not_test_tc_note_unchanged():
#     bind_cmd = "bind -N 'This is a note'  a  display 'that was a'"
#     tc = tc_env()
#     tc.define_tmux_vers(vers="3.2")
#     assert tc.filter_note(bind_cmd) == [bind_cmd]


# def not_test_tc_note_as_comment():
#     #  Also ensure spaces > 2 after note are disgarded
#     tc = tc_env()
#     tc.define_tmux_vers(vers="2.8")
#     assert tc.filter_note("bind -N 'This is a note'    a display 'that was a'") == [
#         "",
#         "#  This is a note",
#         "bind  a display 'that was a'",
#     ]


# def not_test_tc_note_removed():
#     tc = tc_env()
#     tc.define_tmux_vers(vers="2.8")
#     tc.use_notes_as_comments = False
#     assert tc.filter_note("bind -N 'This is a note'  a  display 'that was a'") == [
#         "bind  a  display 'that was a'"
#     ]


# def not_test_tc_note_ends_with_N():
#     #  Ensure empty -N is removed
#     tc = tc_env()
#     tc.define_tmux_vers(vers="2.8")
#     assert tc.filter_note("dummy line -N") == ["dummy line"]


# def not_test_tc_note_single_word():
#     #  needed for coverage
#     tc = tc_env()
#     tc.define_tmux_vers(vers="2.8")
#     tc.filter_note("dummy line -N notice")


# def not_test_tc_note_extra_N_is_purged():
#     #  Ensure extra -N is disgarded
#     note = "missing feature"
#     cmd = "dummy line"
#     cmd_rest = 'M display "mouse toggle needs 2.1"'
#     line = f'{cmd} -N      "{note}" {cmd_rest}'
#     tc = tc_env()
#     tc.define_tmux_vers(vers="2.8")
#     assert tc.filter_note(f'{line} -N "another comment"') == [
#         "",
#         f"#  {note}",
#         f"{cmd} {cmd_rest}",
#     ]


#
#  Conf file
#
# def not_test_tc_conf_file_not_replace(capfd):
#     t = TmuxConfig(parse_cmd_line=False, replace_config=False)
#     with pytest.raises(OSError):
#         t.run()
#     out, _ = capfd.readouterr()
#     assert out.find("Do you wish to replace the default config file (y/n)?") > -1


def not_test_tc_conf_file_create_default(capfd):
    def_conf_file = os.path.expanduser("~/.tmux.conf")
    tmp_conf_file = os.path.expanduser("~/.tmux.conf.orig")
    if os.path.exists(def_conf_file):
        if os.path.exists(tmp_conf_file):
            #  Ensure we start with an empty file
            os.remove(tmp_conf_file)
        shutil.move(def_conf_file, tmp_conf_file)

    t = TmuxConfig(parse_cmd_line=False)
    with pytest.raises(OSError):
        t.run()
    out, _ = capfd.readouterr()
    if os.path.exists(tmp_conf_file):
        shutil.move(tmp_conf_file, def_conf_file)
    assert out.find(
        "Do you wish to create a default config file (y/n)?") > -1


def not_test_tc_conf_file_no_write(capfd):
    if os.path.isdir("/proc/ish"):
        #
        #  iSH doesn't have propper file privs, so will allow user writes
        #  to read only paths
        #
        pytest.skip("skipping test not working on iSH")

    cf_file = "/var/root/foo/tmux.conf"
    with pytest.raises(OSError):
        t = TmuxConfig(parse_cmd_line=False, conf_file=cf_file)
        t.run()
    out, _ = capfd.readouterr()
    assert (
        out.find("ERROR: Could not create directory for config file!") > -1
        or out.find("ERROR: Could not write to config file") > -1
    )


def not_test_tc_conf_file_not_replace_other_conf(capfd):
    t = TmuxConfig(parse_cmd_line=False, conf_file=CONF_FILE)
    with pytest.raises(OSError):
        t.run()
    out, _ = capfd.readouterr()
    assert out.find(f"Do you wish to replace {CONF_FILE} (y/n)?") > -1


def not_test_tc_conf_file_not_embedded():
    tc = NotEmbedded(parse_cmd_line=False, conf_file=CONF_FILE)
    tc.conf_file_header()
    with open(os.path.expanduser(CONF_FILE), "r", encoding="utf-8") as f:
        line = f.readline()  # Just get 1st line
    assert line == "#\n"
    tc.remove_conf_file()


def not_test_tc_conf_file_is_embedded():
    tc = tc_env()
    tc.conf_file_header()
    with open(os.path.expanduser(CONF_FILE), mode="r", encoding="utf-8") as f:
        line = f.readline()  # Just get 1st line
    assert line.find("EMBEDDED-SCRIPTS-STARTING-POINT") > -1
    tc.remove_conf_file()


def not_test_tc_conf_file_header_and_content():
    """Check that TMUX_CONF & TMUX_BIN are correctly set"""
    fname = os.path.expanduser(CONF_FILE)
    tc = NotEmbedded(parse_cmd_line=False, conf_file=CONF_FILE)
    tc.replace_config = True
    tc.conf_file_header()
    tc.run()
    with open(os.path.expanduser(fname), encoding="utf-8") as f:
        lines = f.readlines()
    tmux_bin_found = False
    edit_conf_found = False
    for line in lines:
        if line.find("TMUX_CONF=") > -1:
            conf_file = line.split("=")[1].strip()
            assert conf_file == f'"{fname}"'
        if line.find("TMUX_BIN=") > -1:
            tmux_bin = line.split("=")[1].strip()
            assert tmux_bin == f'"{tc.tmux_bin}"'
            tmux_bin_found = True
        if line.find("Edit config") > -1:
            edit_conf_found = True
        if line.find(CONTENT_LINE) > -1:
            # This should come after the TMUX_BIN line
            # and before the "Edit config" line
            assert tmux_bin_found is True
            assert edit_conf_found is not True

    assert edit_conf_found is True
    tc.remove_conf_file()


def my_func():
    # This function does not raise any exceptions
    pass


#
#  tmux bin versions
#
def not_test_tc_default_bin_found():
    # with pytest.raises(Exception):
    #     TmuxConfig(parse_cmd_line=False, conf_file=CONF_FILE)
    if pytest.raises(Exception):
        # Use the pytest.raises context manager to capture
        # any exceptions raised by the function
        my_func()
    assert True


def not_test_tc_bin_not_found():
    bad_tmux = "tmmuxx"
    with pytest.raises(TmuxConfNotTmuxCommand):
        TmuxConfig(parse_cmd_line=False, conf_file=CONF_FILE, tmux_bin=bad_tmux)


def not_test_tc_bin_empty():
    t = TmuxConfig(parse_cmd_line=False, conf_file=CONF_FILE)
    with pytest.raises(TmuxConfNotTmuxCommand):
        t.use_tmux_bin("")


def not_test_tc_bin_incorrect():
    wrong_bin = "python3"  # must exist even if not tmux...
    t = TmuxConfig(parse_cmd_line=False, conf_file=CONF_FILE)
    with pytest.raises(TmuxConfNotTmuxCommand):
        t.use_tmux_bin(wrong_bin)


# def not_test_tc_bin_auto_find():
#     #  coverage
#     t = TmuxConfig(parse_cmd_line=False, conf_file=CONF_FILE)
#     t.define_tmux_vers(tmux_bin="")
#     t.find_cmd_2()
#     t.find_cmd_3()


def not_test_tc_non_default_version():
    other_version = "4.0"
    remove_conf_file()

    t = PluginsSample(
        parse_cmd_line=False,
        conf_file=CONF_FILE,
        tmux_version=other_version,
        plugins_display=3,
    )
    with pytest.raises(SystemExit):
        t.run()


def not_test_tc_too_old_for_plugins(capfd):
    ancient_vers = "1.0"
    with pytest.raises(SystemExit) as exc:
        TmuxConfig(parse_cmd_line=False, tmux_version=ancient_vers, plugins_display=1)
        # a = 2
    out, _ = capfd.readouterr()
    assert exc.value.code.find("Versions < 1.9 does not support tpm") > -1


#
#  Working with plugins
#
def not_test_tc_plugin_found():
    t = prep_plugin_class(cls=PluginsSample, version="2.4")
    assert t.plugins.found() == [PLUGIN_NAME]
    assert t.plugins.found(short_name=False) == [f"{PLUGIN_SOURCE}/{PLUGIN_NAME}"]


def not_test_tc_plugin_write():
    t = prep_plugin_class(cls=PluginsSample)
    t.run()
    install_plugins()


def not_test_tc_plugin_display_2():
    t = prep_plugin_class(cls=PluginsSample, plugins_display=2)
    with pytest.raises(SystemExit):
        t.run()
    # out, _ = capfd.readouterr()
    # print(f">> out [{out}]")
    # assert 1 == 2


def not_test_tc_plugin_display_3():
    t = prep_plugin_class(cls=PluginsSample, plugins_display=3)
    with pytest.raises(SystemExit):
        t.run()


def not_test_tc_plugin_unavailable():
    t = prep_plugin_class(cls=PluginsSample, version="2.0")
    assert not t.plugins.found()
    assert not t.plugins.found(short_name=False)


def not_test_tc_plugins_disabled():
    t = prep_plugin_class(cls=PluginsDisabled, version="2.4")
    assert not t.plugins.found()
    assert not t.plugins.found(short_name=False)


def not_test_tc_plugin_duplicate():
    with pytest.raises(SystemExit):
        prep_plugin_class(cls=PluginsDuplicated, version="2.4")


#
#  Checks that only plugins of matching version are selected
#
# @pytest.mark.parametrize(
#     "vers, plugins_expected",
#     [
#         (2.2, 0),
#         (2.8, 1),
#         (3.2, 2),
#     ],
# )
# def not_test_tc_plugins_display_info(vers, plugins_expected, capsys):
#     ps = prep_plugin_class(cls=PluginsSample, version=vers)
#     #  to test both verbose and not for plugin display in order to
#     #  increase coverage
#     if plugins_expected == 2:
#         ps.plugins._plugins_display = 3
#     elif plugins_expected == 1:
#         ps.plugins._plugins_display = 2
#     try:
#         ps.plugins.display_info()
#     except SystemExit:
#         #  display_info() terminates prog after displaying info,
#         #  this forces continuation
#         pass
#     captured = capsys.readouterr()
#     do_count = False
#     plugins_found = 0
#     for line in captured.out.split("\n"):
#         print(line)
#         if line.find("Plugins ignored") > -1 or line.find("Unused plugins found") > -1:
#             #  Stop counting when ignored list starts
#             do_count = False
#         if do_count and line.find(" - ") > -1:
#             plugins_found += 1
#         if line.find("|  Min") > -1:
#             #  Start counting after header before used plugins
#             do_count = True
#     assert plugins_found == plugins_expected


# @pytest.mark.parametrize(
#     "plugin_cls",
#     [
#         (PluginsSample),
#         (PluginsManual),
#     ],
# )

# def not_test_tc_plugins_parse(plugin_cls):
#     ps = prep_plugin_class(cls=plugin_cls)
#     output = ps.plugins.parse()
#     assert isinstance(output, list)


#
#  tmate tests
#
def not_test_tc_tmate():
    if not (tmate_cmd := run_shell("command -v tmate")):
        pytest.skip("tmate not found")
    TmuxConfig(parse_cmd_line=False, conf_file="~/.tmux.conf", tmux_bin=tmate_cmd)


# def not_test_tc_is_not_tmate():
#     tc = tc_env()
#     assert tc.is_tmate() is not True


# def not_test_tc_is_tmate():
#     tc = tc_env()
#     #  This assumes there is a tmate in PATH
#     if not (tmate_cmd := run_shell("command -v tmate")):
#         pytest.skip("tmate not found")
#     tc.use_tmux_bin(tmate_cmd)
#     assert tc.is_tmate() is True


def not_test_tc_tmate_get_env():
    tc = tc_env(conf_file="~/.tmate.conf")
    #  This assumes there is a tmate in PATH
    if not (tmate_cmd := run_shell("command -v tmate")):
        return  # Can't do this test
    tc.use_tmux_bin(tmate_cmd)
    plugins_dir, _ = tc.plugins.get_env()
    assert plugins_dir == f"{os.getenv('HOME')}/.tmate/plugins"



import os

import pytest

# from src.tmux_conf.constants import XDG_CONFIG_HOME
from src.tmux_conf.embedded_scripts import EmbeddedScripts
from src.tmux_conf.plugins import Plugins
from src.tmux_conf.tmux_conf import TmuxConfig
from src.tmux_conf.vers_check import VersionCheck

from .utils_test import tmux_conf_instance

# from .utils_test import tst_conf_file  # CONF_FILE,
# from .utils_test import tst_plugins_dir  # PLUGINS_DIR,
# from .utils_test import install_plugins, tmp_conf_base, tmux_conf_instance


class PluginUnsafePluginDir(Plugins):
    def get_env(self):
        plugins_dir = os.path.expanduser("~")
        tpm_env = ""
        return plugins_dir, tpm_env


class Using2Plugins(TmuxConfig):
    plugin_handler = "manual"

    def content(self):
        """Needs an override"""

    def plugin_session_wizard(self):
        return ["27medkamal/tmux-session-wizard", 1, ""]

    def plugin_better_mouse_mode(self):
        return ["jaclu/tmux-better-mouse-mode", 99, ""]

    def plugin_no_slash(self):
        return ["no_slash", 1, ""]


class Using3Plugins(Using2Plugins):
    def plugin_prefix_highlight(self):
        return ["jaclu/tmux-prefix-highlight", 0, ""]


class DuplicatePlugins(Using2Plugins):
    def plugin_better_mouse_mode2(self):
        return ["jaclu/tmux-better-mouse-mode", 99, ""]


def plugins_env(conf_file, plugins_display=0):
    # f"{tmp_path}/tmux/tmux.conf"
    vc = VersionCheck(3.0)
    es = EmbeddedScripts(conf_file=conf_file, use_embedded_scripts=True)
    plugins = Plugins(
        conf_file=conf_file,
        vers_class=vc,
        es_class=es,
        plugins_display=plugins_display,
    )
    return plugins


# def not_test_p_bad_display_param(tmp_path):
#     # tst_conf_file = f"{tmp_path}/tmux/tmux.conf"
#     vc = VersionCheck(3.0)
#     es = EmbeddedScripts(conf_file=tst_conf_file, use_embedded_scripts=True)
#     with pytest.raises(ValueError):
#         Plugins(conf_file=tst_conf_file, vers_class=vc, es_class=es, plugins_display=4)


# def not_test_p_deploy_dir(tst_conf_file):
#     plugins = plugins_env(tst_conf_file)
#     ddir = plugins.get_plugin_dir()
#     assert ddir.endswith("/plugins") is True


# def not_test_p_limited_host_on(tst_conf_file):
#     plugins = plugins_env(tst_conf_file)
#     assert plugins.set_limited_host(True) is True


# def not_test_p_limited_host_off(tst_conf_file):
#     plugins = plugins_env(tst_conf_file)
#     assert plugins.set_limited_host(False) is not True


# def not_test_p_get_env(tst_conf_file):
#     plugins = plugins_env(tst_conf_file)  # "~/.tmux.conf")
#     ddir, _ = plugins.get_env()
#     assert ddir == plugins.get_plugin_dir()


# def not_test_p_get_plugin_dir_tmate():
#     tc = plugins_env(conf_file="~/.tmate.conf")
#     plugin_dir = tc.get_plugin_dir()
#     assert plugin_dir == f"{os.getenv('HOME')}/.tmate/plugins"


# def not_test_p_get_plugin_dir_xdg(tmp_path):
#     tmp_conf_base = f"{tmp_path}/tmux"
#     xdg_orig = os.environ.get(XDG_CONFIG_HOME, "")
#     xdg_location = tmp_conf_base
#     os.environ[XDG_CONFIG_HOME] = xdg_location
#     plugins = plugins_env(f"{tmp_conf_base}/.tmux.conf")
#     plugin_dir = plugins.get_plugin_dir()
#     if xdg_orig:
#         os.environ[XDG_CONFIG_HOME] = xdg_orig
#     else:
#         del os.environ[XDG_CONFIG_HOME]
#     assert plugin_dir.find(xdg_location) > -1


# def not_test_p_manual_handler(tst_conf_file):
#     plugins = plugins_env(tst_conf_file)
#     cont = plugins.mkscript_manual_deploy()
#     assert cont[0].find("Manual Plugin Handling") > -1


# def not_test_p_tpm_handler(tst_conf_file):
#     plugins = plugins_env(tst_conf_file)
#     cont = plugins.mkscript_tpm_deploy()
#     assert cont[0].find("Tmux Plugin Manager") > -1


def test_p_parse(tst_conf_file):
    """Do a parse with no plugins defined"""
    plugins = plugins_env(tst_conf_file)
    assert not plugins.parse()


def test_p_def_conf_disp_info():
    plugins = plugins_env("~/.tmux.conf", plugins_display=1)
    with pytest.raises(SystemExit):
        plugins.display_info()


# def bad_test_p_install(tst_conf_base, tst_conf_file, tst_plugins_dir):
#     install_plugins(tst_conf_file)  # Manually install all plugins


# def not_test_p_clear_plugins(tst_conf_base, tst_conf_file, tst_plugins_dir):
#     cls = Using2Plugins

#     #  Create a dummy dir, to ensure clear_plugins work as intended
#     os.makedirs(f"{tst_conf_base}/foo", exist_ok=True)

#     #  First clear all plugins installed
#     tmux_conf_instance(cls, tst_conf_file, clear_plugins=True).run()

#     # Ensure plugins dir is empty
#     ddir = os.listdir(tst_plugins_dir)
#     if len(ddir) > 0:
#         raise ValueError("Plugin dir not empty!")

#     # vc = VersionCheck(3.0)
#     # es = EmbeddedScripts(conf_file=tst_conf_file, use_embedded_scripts=True)
#     # Plugins(conf_file=tst_conf_file, vers_class=vc, es_class=es, clear_plugins=True)
#     # # no exception is good enough for passing


# # def bad_test_p_installAgain(tmp_conf_base):
# #     """Install again to ensure plugins are present for further tests"""
# #     test_p_install(tmp_conf_base, cls=Using3Plugins)


def test_p_display(capfd, tst_conf_file):
    # tmux_conf_instance(cls, tst_conf_file, clear_plugins=True).run()
    t = tmux_conf_instance(Using2Plugins, tst_conf_file, plugins_display=2)
    with pytest.raises(SystemExit):
        t.run()

    out, _ = capfd.readouterr()
    # print(f">> out [{out}]")
    plugins_not_installed = []
    plugins_installed = []
    plugins_ignored = []
    gather_listed = gather_ignored = False
    for line in out.split("\n"):
        # print(f">> line [{line}]")
        if line.find("|  Min version") > -1:
            gather_listed = True
            # print(">> activating gather on line", line)
            continue
        if gather_listed:
            if not line.strip():
                gather_listed = False
                # print(">> canceled gather on line", line)
                continue
            plugin_name = line.split()[0]
            if line.find("Not installed") > -1:
                plugins_not_installed.append(plugin_name)
            else:
                plugins_installed.append(plugin_name)
        if line.find("vers|") == 0:
            gather_ignored = True
            continue
        if gather_ignored and line:
            plugin_name = line.split()[-1]
            plugins_ignored.append(plugin_name)

    # print(f">> installed     [{plugins_installed}]")
    # print(f">> not installed [{plugins_not_installed}]")
    # print(f">> ignored [{plugins_ignored}]")
    assert len(plugins_installed) == 1
    assert len(plugins_not_installed) == 1
    assert len(plugins_ignored) == 1


# def not_test_p_display3(capfd):
#     t = tmux_conf_instance(Using2Plugins, plugins_display=3)
#     with pytest.raises(SystemExit):
#         t.run()


# def not_test_p_duplicate_plugin(capfd):
#     with pytest.raises(SystemExit):
#         tmux_conf_instance(DuplicatePlugins).run()
#     out, _ = capfd.readouterr()
#     assert out.find("defined more than once") > -1


# def not_test_p_limited_host():
#     t = tmux_conf_instance(Using2Plugins)
#     t.is_limited_host = True
#     try:
#         t.run()
#     except SystemExit:
#         pass
#     t.plugins.mkscript_tpm_deploy()


# def not_test_p_suspicious_clear():
#     t = tmux_conf_instance(Using2Plugins)
#     t.plugins = PluginUnsafePluginDir(
#         tst_conf_file,
#         vers_class=t.vers,
#         es_class=t.es,
#         plugin_handler=t.plugin_handler,
#         clear_plugins=False,
#         plugins_display=0,
#     )
#     with pytest.raises(FileNotFoundError):
#         t.plugins.clear()


# def not_test_p_missing_plugin_dir():
#     """Security checks to minimize risk of harmfull deletion"""
#     if not tst_plugins_dir:
#         raise SystemError("PLUGINS_DIR undefined")
#     wrong_dir = "/moved_plugins"
#     parts = "/".join(tst_plugins_dir.split("/")[:-1])
#     target = parts + wrong_dir
#     shutil.move(tst_plugins_dir, target)
#     t = tmux_conf_instance(Using2Plugins)
#     t.plugins.clear()
#     if target.find(wrong_dir) != len(parts):
#         raise ValueError(
#             f"Not expected target [{target}] [{len(parts)} [{target.find(wrong_dir)}]"
#         )
#     shutil.rmtree(target)



import os
import tempfile
from pathlib import Path
from stat import S_IREAD, S_IWRITE

import pytest
from src.tmux_conf import utils

from .utils_test import CONF_FILE


@pytest.mark.parametrize(
    "test_input,expected",
    [
        ("no bticks", False),
        ("bticks present ` ", True),
        ("double escaped btick \\` ", False),
    ],
)
def not_test_utils_btick(test_input, expected):
    assert utils.btick_unescaped(test_input) is expected


def not_test_utils_expanduser_plus():
    username = str(os.environ.get("USER"))
    tilde = utils.expanduser_plus("~")
    home = utils.expanduser_plus("$HOME")
    assert tilde == home
    assert home.find(username) > -1
    assert tilde.find(username) > -1


@pytest.mark.parametrize(
    "test_input,expected",
    [
        ("/etc/passwd", False),
        ("/bin/sh", True),
    ],
)
def not_test_utils_is_executable(test_input, expected):
    assert utils.is_executable(test_input) is expected


@pytest.mark.parametrize(
    "test_input",
    [
        ("-c"),
        ("--clear_plugins"),
    ],
)
def not_test_utils_parse_clear_plugins(test_input):
    args = utils.parse_cmdline([test_input])
    assert args.clear_plugins is True


def not_test_utils_parse_replace():
    args = utils.parse_cmdline(["-r"])
    assert args.replace is True
    args = utils.parse_cmdline(["--replace"])
    assert args.replace is True


@pytest.mark.parametrize(
    "test_input,expected",
    [
        ("", 0),
        (0, 0),
        (1, 1),
        (2, 2),
        (3, 3),
        (4, 0),
    ],
)
def not_test_utils_parse_plugins_display(test_input, expected):
    if expected > 0:
        args = utils.parse_cmdline(["-p", f"{test_input}"])
        assert args.plugins_display == expected
        args = utils.parse_cmdline(["--plugins_display", f"{test_input}"])
        assert args.plugins_display == expected
    else:
        with pytest.raises(SystemExit):
            args = utils.parse_cmdline(["-p", f"{test_input}"])
        with pytest.raises(SystemExit):
            args = utils.parse_cmdline(["--plugins_display", f"{test_input}"])


def not_test_utils_parse_tmux_bin():
    dummy_bin = "foo"
    args = utils.parse_cmdline(["-t", dummy_bin])
    assert args.tmux_bin == dummy_bin
    args = utils.parse_cmdline(["--tmux_bin", dummy_bin])
    assert args.tmux_bin == dummy_bin


def not_test_utils_parse_forced_version():
    args = utils.parse_cmdline(["-V", "2.4"])
    assert args.forced_version == "2.4"
    args = utils.parse_cmdline(["--forced-version", "2.4"])
    assert args.forced_version == "2.4"


def not_test_utils_parse_conf_file():
    args = utils.parse_cmdline([CONF_FILE])
    assert args.conf_file == CONF_FILE


def not_test_utils_run_shell():
    hw = "Hello world"
    assert utils.run_shell(f"echo {hw}") == hw


def not_test_utils_tilde_home_dir():
    tilde_name = "~/tmp"
    full_path = utils.expanduser_plus(tilde_name)
    assert utils.tilde_home_dir(full_path) == tilde_name


def not_test_utils_not_tilde_home_dir():
    tmp_dir = tempfile.mkdtemp()
    assert utils.tilde_home_dir(tmp_dir) == tmp_dir


def not_test_utils_conf_file_useable_empty():
    with pytest.raises(SyntaxError):
        utils.verify_conf_file_usable("")


def not_test_utils_conf_file_useable_relative():
    #
    #  Test that if no path is given, it is expanded to cwd
    #
    conf_relative = "local_conf.conf"
    conf_file = os.path.join(os.getcwd(), conf_relative)
    generated_full_path_cf = utils.verify_conf_file_usable(conf_relative)
    assert generated_full_path_cf == conf_file


def not_test_utils_conf_file_useable_readonly_fs():
    conf_ro_fs = "/root/foo/dummy.conf"
    with pytest.raises(OSError):
        utils.verify_conf_file_usable(conf_ro_fs)


def not_test_utils_conf_file_useable_readonly_file():
    conf_ro_file = tempfile.NamedTemporaryFile()
    Path(conf_ro_file.name).touch()
    os.chmod(conf_ro_file.name, S_IREAD)
    with pytest.raises(IOError):
        utils.verify_conf_file_usable(conf_ro_file.name)
    os.chmod(conf_ro_file.name, S_IWRITE)



import os

import pytest

from src.tmux_conf.embedded_scripts import EmbeddedScripts
from src.tmux_conf.utils import run_shell

from .utils_test import CONF_FILE

SCRIPT_NAME = "foo123"


def es_env(conf_file=CONF_FILE, use_embedded_scripts=True):
    return EmbeddedScripts(
        conf_file=conf_file, use_embedded_scripts=use_embedded_scripts
    )


def es_hello_world(conf_file=CONF_FILE, use_embedded_scripts=True, use_bash=False):
    es = es_env(conf_file=conf_file, use_embedded_scripts=use_embedded_scripts)
    sh = [
        f"""
{SCRIPT_NAME}() {{

    echo "Hello world"

}}

"""
    ]
    sh.append("")
    es.create(SCRIPT_NAME, sh, use_bash=use_bash)
    return es


def get_shebang(fname):
    with open(fname, encoding="utf-8") as f:
        line = f.readline()  # Just get 1st line
    return line.strip()


def not_test_es_no_plugins_should_return_empty():
    es = es_env()
    assert isinstance(es.content(), list)


def not_test_es_relative_conf_file():
    conf_file = "dummy_conf"
    es = es_env(conf_file)
    # pylint: disable=W0212
    print(f">> conf_file: {es._conf_file}")
    assert es._conf_file.split("/")[-1] == conf_file


def not_test_es_xdg():
    if not run_shell("command -v bash"):
        pytest.skip("No bash found")
    os.environ["XDG_CONFIG_HOME"] = "/tmp"  # nosec
    es = es_hello_world(use_embedded_scripts=False, use_bash=True)
    bash_sh = es.run_it(SCRIPT_NAME).split('"')[1]
    os.environ.pop("XDG_CONFIG_HOME")
    # Simplify by not caring about path to bash
    assert get_shebang(bash_sh) == "#!/usr/bin/env bash"


def not_test_es_create_external_bash():
    if not run_shell("command -v bash"):
        pytest.skip("No bash found")
    es = es_hello_world(use_embedded_scripts=False, use_bash=True)
    bash_sh = es.run_it(SCRIPT_NAME).split('"')[1]
    # Simplify by not caring about path to bash
    assert get_shebang(bash_sh) == "#!/usr/bin/env bash"


def not_test_es_create_external_sh():
    es = es_hello_world(use_embedded_scripts=False)
    shell_used = es.run_it(SCRIPT_NAME).split('"')[1]
    assert get_shebang(shell_used) == "#!/bin/sh"


def not_test_es_run_it_posix():
    es = es_hello_world()
    assert es.run_it(SCRIPT_NAME) == f'run "cut -c3- {CONF_FILE} | sh -s {SCRIPT_NAME}"'


def not_test_es_run_it_bash():
    if not run_shell("command -v bash"):
        pytest.skip("No bash found")

    es = es_hello_world(use_bash=True)
    s = es.run_it(SCRIPT_NAME)
    # Simplify by not caring about path to bash
    assert (
        s.startswith(f'run "cut -c3- {CONF_FILE} | ') is True
        and s.endswith(f'/bash -s {SCRIPT_NAME}"') is True
    )


def not_test_es_run_it_in_bg():
    es = es_hello_world()
    assert (
        es.run_it(SCRIPT_NAME, in_bg=True)
        == f'run -b "cut -c3- {CONF_FILE} | sh -s {SCRIPT_NAME}"'
    )


def not_test_es_content():
    es = es_hello_world()
    cont = es.content()
    errors = []
    if cont[1] != f"# {SCRIPT_NAME}() {{":
        errors.append(f"Bad script name line: {cont[1]}")
    if cont[3] != '#     echo "Hello world"':
        errors.append(f"Bad echo: {cont[2]}")
    assert not errors, "errors occured:\n{}".format("\n".join(errors))


def not_test_es_get_dir_embedded():
    es = es_hello_world()
    with pytest.raises(SyntaxError):
        es.get_dir()


def not_test_es_get_dir_external():
    es = es_hello_world(use_embedded_scripts=False)
    d = es.get_dir()
    assert d.endswith("/scripts")


def not_test_es_get_dir_external_def_conf():
    es = es_env(conf_file="~/.tmux.conf", use_embedded_scripts=False)
    d = es.get_dir()
    assert d == f"{os.getenv('HOME')}/.tmux/scripts"


def not_test_es_get_dir_external_rel_conf():
    es = es_env(conf_file="tmux.conf", use_embedded_scripts=False)
    d = es.get_dir()
    assert d.endswith("/tmux/scripts")



#  -*- mode: python; mode: fold -*-
#
#  Copyright (c) 2022-2024: Jacob.Lundqvist@gmail.com
#  License: MIT
#
#  Part of https://github.com/jaclu/tmux-conf
#

"""base packet imports"""

from .constants import __version__  # noqa: F401
from .tmux_conf import TmuxConfig

__all__ = ["TmuxConfig"]



#
#  Copyright (c) 2022-2024: Jacob.Lundqvist@gmail.com
#  License: MIT
#
#  Part of https://github.com/jaclu/tmux-conf
#
#  See constants.py for version info
#
#  See the README.md in the repository for more info
#

""" Class that handles tmux plugins """

import os
import shutil
import sys
from dataclasses import dataclass
from typing import Callable, Dict, Tuple

import __main__

from .constants import XDG_CONFIG_HOME
from .embedded_scripts import EmbeddedScripts
from .vers_check import VersionCheck

PLUGIN_VERS_MIN = 0
PLUGIN_MTHD = 1
PLUGIN_STATIC_CODE = 2

# str, Tuple[str, Callable[[], list[str]], str]


@dataclass
class Plugin:
    vers_min: str
    plug_def: Callable[[bool], bool]
    code: str


#
#  Tasks related to plugins, taken separately to keep files smaller
#
class Plugins:
    """Class that handles tmux plugins"""

    def __init__(
        self,
        conf_file: str,  # must use tilde for home dir
        vers_class: VersionCheck,
        es_class: EmbeddedScripts,
        plugin_handler: str = "tmux-plugins/tpm",
        clear_plugins: bool = False,
        plugins_display: int = 0,
    ):
        self._conf_file = conf_file
        self._vers = vers_class
        self._es = es_class
        self._plugin_handler = plugin_handler

        self._is_limited_host = False

        self._used_plugins: Dict[str, Tuple[str, Callable[[], list[str]], str]] = {}

        # plugins incompatible with this version
        self._skipped_plugins: list[Tuple[str, str]] = []

        self._fnc_activate_tpm = "activate_tpm"
        self._fnc_activate_manually = "activate_plugins_mamually"

        if plugins_display not in [0, 1, 2, 3]:
            raise ValueError("plugins_display must be one of: 0, 1, 2, 3")
        self._plugins_display = plugins_display
        if clear_plugins:
            #
            #  if plugins_display is set, it will terminate the app,
            #  so lets do this before
            #
            self.clear()

    # def defined(self, short_name: bool = True):
    #     """Returns a list of plugin names being defined.
    #     This also includes defined but not installed plugins.
    #     """
    #     return

    def found(self, short_name: bool = True):
        """OBSOLETE! replaced ny installed()
        TODO: Remove this function!
        Returns a list of plugin names being used.
        If short_name is False will return the full name including
        source, this is needed when cloning the repo, but less desired
        when just checking if a given plugin is used in most cases.
        Since if a different fork of it is being used, the name would
        not match."""
        return self.installed(short_name=short_name)

    def installed(self, short_name: bool = True):
        """Returns a list of plugin names being installed.
        This excludes defined but not already installed plugins.
        Purpose is when defining the status bars, in order not to define
        entries for absent plugins

        Returns a list of plugin names being used.
        If short_name is False will return the full name including
        source, this is needed when cloning the repo, but less desired
        when just checking if a given plugin is used in most cases.
        Since if a different fork of it is being used, the name would
        not match."""
        result = []
        for full_name in self._used_plugins:
            name = full_name
            if short_name:
                # Try to return only actual plugin name, without provider
                if full_name.find("/") > -1:
                    name = full_name.split("/")[1]
            result.append(name)
        return result

    def get_plugin_dir(self) -> str:
        """Returns dir where plugins will be installed."""
        plugins_dir, _ = self.get_env()
        return plugins_dir

    # ================================================================
    #
    #        Rest is mostly for internal usage
    #
    # ================================================================

    def set_limited_host(self, is_limited: bool) -> bool:
        """For this module, this indicates that the host is considered
        slowish, so extended status updates should be provided during
        plugin init, to help indicating when all the plugins are setup.
        Om normalish system, this is close to instantaneous, so step by
        step progress is not really meaningful.
        """
        self._is_limited_host = is_limited
        return self._is_limited_host

    def scan(self, plugin_methods):  # list[Callable[[], list[str]]]) -> None:
        """Investigate all defined plugin methods, and determine if a
        given plugin can be used depending on running tmux"""
        duplicate_check = []
        for plugin_mthd in plugin_methods:
            plugin_name, vers_min, code = plugin_mthd()
            if plugin_name in duplicate_check:
                print(f'ERROR: plugin "{plugin_name}" defined more than once:')
                sys.exit(1)
            else:
                duplicate_check.append(plugin_name)
            #  Since plugi method might define vers_min as a float
            #  we need to convert it to a string both for used and ignored plugins
            if self._vers.is_ok(vers_min):
                self._used_plugins[plugin_name] = (
                    str(vers_min),  # PLUGIN_VERS_MIN
                    plugin_mthd,  # PLUGIN_MTHD
                    code,  # PLUGIN_STATIC_CODE
                )
            else:
                self._skipped_plugins.append((str(vers_min), plugin_name))
        self._skipped_plugins.sort()

    def display_info(self) -> str:
        """List selected and ignored plugins, depending on param"""
        print(f"\n\t=====  tmux {self._vers.get()} - Plugins defined  =====")
        print(f" for: {__main__.__file__}")

        #
        #  First loop to find longest plugin name
        #
        max_l_name = 0
        for name in self._used_plugins:
            max_l_name = max(max_l_name, len(name) + 2)

        if self._used_plugins:
            print("\n\t-----   Plugins used   -----")
            print(f'{"Plugin":<{max_l_name}}|  Min version')

        #
        #  Create list of items in plugins dir
        #
        # plugin_items = next(os.walk(self.get_plugin_dir()))[1]
        plugin_items = []
        plugin_dir = self.get_plugin_dir()
        if os.path.exists(plugin_dir):
            for f in os.scandir(plugin_dir):
                if f.is_dir():
                    plugin_items.append((f.path.split("/")[-1]))
        _ = self._remove_if_found(plugin_items, "tpm")

        #
        #  When exec'ing the plugin code below, write output to stdout,
        #  not to config file. Since program will exit after displaying
        #  plugin details, no need to reset this variable.
        #
        verbose = self._plugins_display == 3

        for name, info in self._used_plugins.items():
            if verbose:
                inner_name = self._name_sans_prefix(name)
                suffix = self._remove_if_found(
                    plugin_items, inner_name, " *** Not installed ***"
                )
                print("".ljust(len(inner_name) + 2, "-"))
                print(
                    f"> {inner_name:<{max_l_name - 2}} - "
                    f"{info[PLUGIN_VERS_MIN]} {suffix}"
                )
                info[1]()  # PLUGIN_MTHD
                #
                #  Skip indention, for easier read
                #
                for line in info[2].split("\n"):  # PLUGIN_STATIC_CODE
                    # if line == line.strip():
                    print(f"{line.strip()}")
                # print()
            else:
                inner_name = self._name_sans_prefix(name)
                suffix = self._remove_if_found(
                    plugin_items, inner_name, " *** Not installed ***"
                )
                print(f"{inner_name:<{max_l_name}} - {info[PLUGIN_VERS_MIN]} {suffix}")

        #  Remove skipped plugins from plugin_items
        for _, name in self._skipped_plugins:
            inner_name = self._name_sans_prefix(name)
            _ = self._remove_if_found(plugin_items, inner_name)

        if plugin_items:
            print("\n-----   Unused plugins found   -----")
            for s in plugin_items:
                print("\t", s)

        if not self._skipped_plugins or self._plugins_display != 2:
            sys.exit(0)

        if self._skipped_plugins:
            print()

        #
        #  List all plugins installed, but not used
        #
        max_l_v = 0
        self._skipped_plugins.sort()
        for vers, name in self._skipped_plugins:
            max_l_v = max(max_l_v, len(vers))
        print("-----   Plugins ignored   -----")
        print(f'{"Min":<{max_l_v}}|{" Plugin name":<{max_l_name}}')
        print(f'{"vers":<{max_l_v}}|\n')
        for vers, name in self._skipped_plugins:
            print(f"{vers:>{max_l_v}}  {name:<{max_l_name}}")

        sys.exit(0)

    def parse(self):
        """This package will use any plugin defining methods it can find.
        They will be sorted alphabetically by method name, normally the order
        of plugins do not matter, but if you do want to force the sorting,
        just change the method name accordingly.

        They are assumed to start with plugin_ and have the following format:

        The method should return a triplet:
         min_version,
         plugin name - needed when installing it, and to identify it.
         code snippet defining plugin variables, can be empty string.

        The code snippet will just be copied as is, so if a version or other
        check is needed, put that code before the return. Within the method
        all normal tmux-conf functionality is available.

        Here are a few examples:
        """
        if not self._used_plugins:
            return []

        output = []

        #
        #  First ensure that plugin code that needs to process the
        #  environment is done and output is written to config.
        #
        for name, info in self._used_plugins.items():
            info[1]()

        #
        #  Add plugin references and hard coded plugin settings.
        #
        for name, info in self._used_plugins.items():
            output.append("#------------------------------")
            if self._vers.is_ok("1.8"):
                output.append(f'set -g @plugin "{name}"')
                for line in info[2].split("\n"):
                    output.append(line.strip())
            else:
                #  prior to 1.8, any variables starting wiith @ would get tmux
                #  stuck parsing the config file, so plugins without any such
                #  setting could be handled by activate_plugins_mamually()
                #  This is such a rare edge case that it is not worh handling
                output += [
                    f"# plugin: {name}",
                    "# in versions < 1.8 @variables can not be used",
                    "",
                ]

        if self._plugin_handler == "manual":
            #
            #  Setup manual plugin handling, check for each listed
            #  plugin and install if missing
            #
            output.append(self.mkscript_manual_deploy())
            output.append(self._es.run_it(self._fnc_activate_manually, in_bg=True))
        elif self._plugin_handler:
            #
            #  For any other _plugin_handler setting, assume it is tpm
            #  or a clone.
            #
            output.append(self.mkscript_tpm_deploy())
            output.append(self._es.run_it(self._fnc_activate_tpm, in_bg=True))
        output.append("")  # spacer between sections
        return output

    def get_env(self) -> tuple[str, str]:
        """get environment"""
        location = os.path.dirname(os.path.expanduser(self._conf_file))
        if location == os.path.expanduser("~"):
            #
            #  If conf file is default, plugins have a non standard location
            #
            if self._conf_file.find("tmate") > -1:
                plugins_dir = os.path.expanduser("~/.tmate/plugins")
                tpm_env = ""
            else:
                plugins_dir = os.path.expanduser("~/.tmux/plugins")
                tpm_env = ""
        else:
            xdg_home = os.environ.get(XDG_CONFIG_HOME)
            if xdg_home:
                conf_base = os.path.expanduser(xdg_home)
            else:
                conf_base = os.path.dirname(location)

            plugins_dir = os.path.join(conf_base, "tmux", "plugins")
            tpm_env = os.path.expanduser(f'XDG_CONFIG_HOME="{conf_base}" ')

        return plugins_dir, tpm_env

    def mkscript_manual_deploy(self):
        """This script is run as tmux starts, all non-present
        plugins are installed, and an attempt is done to initialize
        each plugin.
        """
        output = []
        output.append(
            """
        #======================================================
        #
        #   Manual Plugin Handling
        #
        #======================================================
        """
        )
        plugins_dir, _ = self.get_env()
        plugins = ""
        for p in self.found(short_name=False):
            plugins += " " + p

        activate_manually_sh = [
            f"""
#
#  This is a manual plugin handler, does not scan for items.
#  The plugins list must be altered manually
#
{self._fnc_activate_manually}() {{
    mkdir -p "{plugins_dir}"

    plugins=({plugins} )
    for plugin in "${{plugins[@]}}"; do
        name="$(echo "$plugin" | cut -d / -f2)"
        if [[ ! -d "{plugins_dir}/$name" ]]; then
            $TMUX_BIN display "cloning  $name"
            git clone "https://github.com/$plugin" "{plugins_dir}/$name"
        fi
        #  argh, handle softlinks to plugin folders, in order to find init...
        d_plugin_folder="$(realpath "{plugins_dir}/$name")"
        init_script="$(find "$d_plugin_folder" -maxdepth 1 | grep tmux$ | head -n 1)"
        if [[ -n "$init_script" ]]; then
            $TMUX_BIN display "running: $init_script"
            $init_script || $TMUX_BIN display "ERROR in $init_script"
        else
            $TMUX_BIN display "Could not find init for plugin: $name"
            sleep 2
        fi
    done
}}""",
        ]
        self._es.create(
            self._fnc_activate_manually,
            activate_manually_sh,
            use_bash=True,
            built_in=True,
        )
        return output

    def mkscript_tpm_deploy(self):
        """If tpm is present, it is started.
        If not, it is installed and requested to install all
        defined plugins.
        """
        output = []
        output.append(
            """
        #======================================================
        #
        #   Tmux Plugin Manager
        #
        #======================================================
        """
        )
        #  os.makedirs(plugins_dir, exist_ok=True)
        plugins_dir, tpm_env = self.get_env()
        tpm_location = os.path.join(plugins_dir, "tpm")
        tpm_app = os.path.join(tpm_location, "tpm")
        activate_tpm_sh = [
            f"""
{self._fnc_activate_tpm}() {{
    #
    #  Initialize already installed tpm if found
    #
    if [ -x "{tpm_app}" ]; then
        {tpm_env}{tpm_app}
        exit 0
    fi

    #  Create plugin dir if needed
    mkdir -p "{plugins_dir}"

    #  Remove potentially broken tpm install
    rm -rf "{tpm_location}"

    $TMUX_BIN display "Cloning {self._plugin_handler} into {tpm_location} ..."
    git clone https://github.com/{self._plugin_handler} "{tpm_location}"
    if [ "$?" -ne 0 ]; then
        echo "Failed to clone tmux plugin handler:"
        echo "  https://github.com/{self._plugin_handler}"
        exit 11
    fi

    $TMUX_BIN display "Running cloned tpm..."
    {tpm_env}"{tpm_app}"
    if [ "$?" -ne 0 ]; then
        echo "Failed to run: {tpm_app}"
        exit 12
    fi

    #
    #  this only triggers plugins install if tpm needed to be installed.
    #  Otherwise installing missing plugins is delegated to tpm.
    #  Default trigger is: <prefix> I
    #
    $TMUX_BIN display "Installing all plugins..."
    {tpm_env}"{tpm_location}/bindings/install_plugins"
    if [ "$?" -ne 0 ]; then
        echo "Failed to run: {tpm_location}/bindings/install_plugins"
        exit 12
    fi

    $TMUX_BIN display "Plugin setup completed"
}}""",
        ]
        self._es.create(self._fnc_activate_tpm, activate_tpm_sh, built_in=True)
        return output

    def clear(self):
        """To minimize risk of some bug causing massive file deletion,
        file path is double checked.
        """
        plugins_dir = self.get_plugin_dir()

        b_suspicious = False
        if "tmux/" not in plugins_dir:
            b_suspicious = True

        if b_suspicious:
            raise FileNotFoundError(
                "Refusing to clear plugins due to suspicious "
                + f"plugin dir: [{plugins_dir}]"
            )

        if not os.path.exists(plugins_dir):
            return  # nothing to clear

        for file_name in os.listdir(plugins_dir):
            path = os.path.join(plugins_dir, file_name)
            print(f"removing plugin {file_name}")
            shutil.rmtree(path)

    def _name_sans_prefix(self, name: str) -> str:
        if name.find("/") > -1:
            name = name.split("/")[1]
        return name

    def _remove_if_found(self, lst: list[str], item: str, warning: str = "") -> str:
        if item in lst:
            lst.remove(item)
            warning = ""
        return warning



#
#  Copyright (c) 2022-2024: Jacob.Lundqvist@gmail.com
#  License: MIT
#
#  Part of https://github.com/jaclu/tmux-conf
#
#  See constants.py for version info
#
#  See the README.md in the repository for more info
#
#  All paths stored internally are done so with ~ expanded
#

# pylint: disable=C0116

"""Class that generates a .tmux.conf"""

# from datetime import datetime
import datetime
import os
import shutil
import sys

import __main__

from .constants import __version__
from .embedded_scripts import EmbeddedScripts
from .exceptions import TmuxConfNotTmuxCommand
from .plugins import Plugins
from .utils import (
    btick_unescaped,
    expanduser_plus,
    is_executable,
    parse_cmdline,
    run_shell,
    verify_conf_file_usable,
)
from .vers_check import VersionCheck


class TmuxConfig:
    """Class that generates a .tmux.conf"""

    #
    #  I have forked a tpm that uses TPM_BIN to use the right tmux
    #  context. If you just use one version of tmux, this does not
    #  matter, but if you experiment with various versions,
    #  this is essential. It ensures tpm and other processes started
    #  inside tmux use the running tmux bin.
    #
    #  My fork also reports more detailed progress on what plugin is
    #  processed
    #
    #  set this to 'jaclu/tpm' if you want to try that one!
    #  set it to "manual" if you dont want to use tpm.
    #  set it to '' if you do not want to use plugins at all.
    #
    plugin_handler: str = "tmux-plugins/tpm"

    #
    #  If true and tmux is < 3.1 thus not supporting -N bind notes
    #  this extracts the note and inserts it before the line as a comment.
    #  If false, the note is just discarded.
    #
    use_notes_as_comments: bool = True

    #
    #  If false, scripts are saved in scripts/ inside the tmux conf dir
    #
    use_embedded_scripts: bool = True

    #
    #  Indicates if this host is low on performance, don't enable
    #  demanding plugins etc, I use this on my iSH nodes.
    #  This must be set during __init__(). It is assumed to already been
    #  set to the intended state when self.run() is done.
    #
    #  The only usage of this in this package is to propagate it to
    #  the plugin handler, indicating as plugins are installed and
    #  activated, since this might take a noticeable time on slower nodes.
    #  It is also avaiable in derived plugin_ methods
    #
    #  Other usages of this is up to implementation classes
    #
    is_limited_host: bool = False

    lib_version: str = __version__

    def __init__(
        self,
        parse_cmd_line: bool = True,
        #
        #  if parse_cmd_line is True all other parameters are ignored
        #
        conf_file: str = "~/.tmux.conf",  # where to store conf file
        tmux_bin: str = "tmux",  # Default binary, if none given
        tmux_version: str = "",
        replace_config: bool = False,  # replace config with no prompt
        clear_plugins: bool = False,  # remove all current plugins
        plugins_display: int = 0,  # Display info about plugins
        # then terminate
    ):
        if parse_cmd_line:
            args = parse_cmdline(sys.argv[1:])
            conf_file = args.conf_file
            tmux_bin = args.tmux_bin
            tmux_version = args.forced_version
            replace_config = args.replace
            clear_plugins = args.clear_plugins
            plugins_display = args.plugins_display

        self.tmux_bin = ""

        #  cant use self.is_tmate() at this point, since self.tmux_bin
        #  is not yet set
        if tmux_bin.find("tmate") > -1:
            #  Set tmate defaults
            if not tmux_version:
                tmux_version = "2.4"
            if conf_file == "~/.tmux.conf":
                #  Override default config
                conf_file = "~/.tmate.conf"
                replace_config = True

            if self.plugin_handler:
                #  tmate only supports manual plugin handling
                self.plugin_handler = "manual"

        self.e_c_has_been_called = False
        self._write_stdout = False

        print(f"Processing: {__main__.__file__}")

        self.conf_file = verify_conf_file_usable(conf_file)

        self.find_tmux_bin(tmux_bin, requested_vers=tmux_version)
        if tmux_version and (tmux_version != self.vers.get()):  # type: ignore
            # self.define_tmux_vers(tmux_version)
            if not self.is_tmate():
                #
                #  Only display off version info when bin is tmux
                #
                print()
                print(
                    "Config has been requested for another version of tmux,"
                    + " than the one used to generate this.\n"
                    + "Since the config file will point to the tmux used, "
                    + "this might cause problems\n"
                    + f"\ttmux vers is:    {self.vers.get_actual()}\n"
                    + f"\trequested vers:  {self.vers.get()}"
                )
                print()
                print("WARNING: Running this config with the current tmux")
                print("         might give some errors!")
                print()
        self.replace_config = replace_config
        self.plugins_display = plugins_display

        self.es = EmbeddedScripts(
            conf_file=self.conf_file,
            vers_class=self.vers,  # type: ignore
            use_embedded_scripts=self.use_embedded_scripts,
        )

        #
        #  If current tmux is too old to handle tpm plugins,
        #  disable plugins entirely
        #
        if not self.vers_ok("1.9") and self.plugin_handler:
            # tpm not available before 1.9, since it was defined, ensure
            # this is set to manual
            self.plugin_handler = "manual"

        self.plugins = Plugins(
            conf_file=self.conf_file,
            vers_class=self.vers,  # type: ignore
            es_class=self.es,
            plugin_handler=self.plugin_handler,
            clear_plugins=clear_plugins,
            plugins_display=plugins_display,
        )
        self.write_enable(
            False
        )  # Allow plugin cheks to run without writung to conf file

    # ================================================================
    #
    #  Needs overloading
    #
    # ================================================================

    def content(self) -> None:  # NotImplementedError
        """This method is expected to use self.write() to generate the
        majority of the config file.
        Plugins can be handled separately (check the README) or if you
        prefer you can just drop your current code doing that here.
        """
        raise NotImplementedError("content() must be defined in sub-class!")

    # ================================================================
    #
    #  Optional overloading
    #
    # ================================================================

    def local_overrides(self) -> None:
        """
        Applies local configuration overrides, executed after all other
        configuration steps. The only exception is status bar customizations,
        which should be handled separately in the `status_bar_customization()` method.

        When overriding this method in a subclass, ensure that you call
        `super().local_overrides()` at the start of your method to preserve
        any configuration overrides defined by parent classes before applying
        additional customizations.
        """
        self.write(
            """
        #======================================================
        #
        #   Local overrides
        #
        #======================================================
        """
        )

    # ================================================================
    #
    #  If you prefer another edit key, call this inside your content()
    #
    #  It will be called after content and use the default edit_key
    #  if it has not already been called.
    #
    # ================================================================

    def edit_config(self, edit_key: str = "e") -> None:
        """Managing to first edit the right file then using it to
        generate a tmux config for the right tmux version, and finally
        sourcing it, requires a lot of special considerations,
        therefore this is a template for doing this.

        This is called without param after self.content() by this class.
        In order to give SubClasses a shot at changing the edit_key,
        it can be called within the content() method.
        If it has already been called, it will abort.

        Override with something else or an empty method if this is not
        desired.
        """
        if self.e_c_has_been_called:
            return
        self.e_c_has_been_called = True

        w = self.write
        w(
            f"""

        # ======================================================
        #
        #   Edit config     <prefix>  {edit_key}
        #
        #  Both Homebrew on linux & using an asdf tmux might deploy tmux
        #  outside the generic shell path, in such cases running tmux from a
        #  script inside an interactive session will fail to find the intended
        #  bin.
        #
        #  For this reason the tmux bin used when defining the config,
        #  and the config file is used here to ensure the intended
        #  tmux is being regenerated, and finally sourced.
        #
        # ======================================================
        """
        )

        #
        #  For now we trust the right python to be found in path
        #  If this is not the one with the tmux-conf lib installed
        #  things will fail. If that turns out to be an issue, I guess
        #  storing the "right" python in the conf file would be a solution.
        #
        py_bin = run_shell("command -v python3")

        w(
            f'bind -N "Edit local config files"  {edit_key}  '
            + 'new-window -n "$TMUX_BIN config" '
            + "\"/bin/sh -c '\\${EDITOR:-vi} "
            + "$TMUX_SOURCE && "
            + f"{py_bin} $TMUX_SOURCE -r $TMUX_CONF && sleep 1 && "
            + "$TMUX_BIN source $TMUX_CONF && "
            + '$TMUX_BIN display \\"$TMUX_CONF sourced\\"\'"'
        )
        w("\n")

    # ================================================================
    #
    #  Some general methods that might be useful
    #
    # ================================================================
    def vers_ok(self, vers) -> bool:
        return self.vers.is_ok(vers)  # type: ignore

    # ================================================================
    #
    #        Rest is mostly for internal usage
    #
    # ================================================================

    def run(self) -> None:
        w = self.write

        #
        #  Inform plugin handler if this is to be treated as a
        #  limited host, displaying progress as tpm is processed etc
        #
        self.plugins.set_limited_host(self.is_limited_host)

        self._write_stdout = True  # Needs to be true during plugin init
        self.write_enable(False)
        if self.plugin_handler:
            self.plugins.scan(self.list_plugin_methods())
            if self.plugins_display == 3:
                self.write_enable(True)
            if self.plugins_display:
                self.plugins.display_info()

        if not self.replace_config:
            self.verify_replace()

        self.conf_file_header()

        self.content()
        #
        #  check the edit_config() header in this class for hints on
        #  how to override the edit key
        #
        self.edit_config()
        if self.plugin_handler and self.plugins.found():
            w(
                """#======================================================
            #
            #   Plugins
            #
            # ======================================================\n"""
            )
            for line in self.plugins.parse():
                if isinstance(line, list):
                    for sub_line in line:
                        self.write(sub_line)
                else:
                    self.write(line)

        self.local_overrides()
        #
        #  Should be called as late as possible, to be able to have
        #  gathered all the intended embedded scripts.
        #
        for line in self.es.content():
            self.write(line)

    def list_plugin_methods(self):  # -> list[Callable[[], list[str]]]:
        """Support for plugins.py, provides a list of all plugin_... methods"""
        plugin_mthds = []
        if self.plugin_handler:
            for item in dir(self):
                if item.find("plugin_") or item == "plugin_handler":
                    continue
                plugin_mthds.append(getattr(self, item))
        return plugin_mthds

    def write_enable(self, state: bool) -> None:
        self._write_enabled = state

    def conf_file_header(self) -> None:
        """Creates the conf-file header & sets a few env variables about
        what tmux and config file is used, so that external commands
        know what to call if needed.
        """
        self.remove_conf_file()
        self._write_stdout = False
        self.write_enable(True)

        print(f"Writing tmux {self.vers.get()} config to {self.conf_file}")

        w = self.write
        if self.use_embedded_scripts:
            w(
                """# : << EMBEDDED-SCRIPTS-STARTING-POINT
            #
            # The above line tells embedded scripts where they start
            # further down in this file"""
            )

        w(
            f"""#
        #  This config was created using
        #      https://github.com/jaclu/tmux-conf
        #
        #      Creation time: {datetime.datetime.now().strftime("%y-%m-%d %H:%M:%S")}
        #          tmux-conf: {self.lib_version}
        #         Created on: {run_shell('hostname').strip()}"""
        )
        if self.vers.get() != self.vers.get_actual():  # type: ignore
            w(f"#     actual version: ({self.vers.get_actual()})")
        w(f"#   For tmux version: {self.vers.get()}")  # type: ignore
        w(
            f"""#
        #
        #  Three env variables defining this instance of tmux:
        #

        #
        #  When you might use various tmux instances or tmux is not in
        #  path, the safe bet is to allways use $TMUX_BIN in a shell, not tmux!
        #
        TMUX_BIN="{self.tmux_bin}"

        #
        #  Helper pointitng to what conf file defines this env
        #  If you want to source the config, especially if the config
        #  might not be in the default location do:
        #    $TMUX_BIN source $TMUX_CONF
        #
        TMUX_CONF="{self.conf_file}"

        #
        #  Since this is a compiled config file, please examine $TMUX_SOURCE
        #  for what generated this. That is the right place for any changes
        #  and to examine comments about what is generated and why.
        #  This .conf file will frequently be over-written!
        #
        TMUX_SOURCE="{__main__.__file__}"
        """
        )

    def write(self, cmd: str = "", eol: str = "\n") -> None:
        """Writes tmux cmds to config file

        Filters out -N "note" statements if not supported, so they can
        always be supplied, regardless of tmux version.

        Can handle single and multi line commands, prints LF as a suffix.
        """
        if not self._write_enabled:
            #
            #  Plugin scans might trigger writes, at that point ignore them
            #
            return

        #
        #  Convert to lines actually to be written
        #
        lines: list[str] = []
        for raw_line in cmd.split("\n"):
            #
            #  Returns a list of lines. If input contained a -N and notes
            #  are not supported by this tmux, the note is extracted and
            #  the following is returned.
            #  self.use_notes_as_comments = True
            #   - empty string, vertical spacer
            #   - note as a comment
            #   - the remainder of the initial line
            #  self.use_notes_as_comments = False
            #   - the remainder of the initial line
            #
            for line in self.filter_note(raw_line.strip()):
                lines.append(line)

        with open(self.conf_file, "a", encoding="utf-8") as f:
            for line in lines:
                if self.use_embedded_scripts and (btick_unescaped(line)):
                    raise SyntaxError(
                        "Un-escaped back-ticks can not be present in "
                        + "the generated config when\n"
                        + "embedded_scripts are used!"
                    )
                f.write(f"{line}{eol}")

    def filter_note(self, line: str):
        """Returns list of lines, if notes are not supported
        first an empty spacer line, then the note as a comment,
        and finally the actual command without the note
        if self.use_notes_as_comments is False
        only the third line from above is returned
        """

        if (
            not line
            or line[0] == "#"
            or line.find("-N") < 0
            or (self.vers_ok("3.1") and self.use_notes_as_comments)
        ):
            return [line]
        parts = line.split("-N")
        pre = parts[0].strip()
        post = parts[1].strip()
        if not post:
            # Propably an -N at end of line, so not related to a note
            return [pre]
        p0 = post[0]
        if p0 in {'"', "'"}:
            end_note = post[1:].find(p0)
            x = end_note + 1
            note = post[1:x]
            y = end_note + 2
            post = post[y:]
        else:  # single word note
            note_end = post.find(" ")
            note = post.split()[0]
            post = post[note_end:]

        while post.find("   ") == 0:
            post = post[1:]
        new_line = pre + post
        if self.use_notes_as_comments:
            return ["", f"#  {note}", new_line]

        return [new_line]

    #
    #  Should conf file be replaced?
    #
    def verify_replace(self) -> None:
        """Get verification that a config should be over-written
        the param replace_config=True skips this check
        """
        default_conf_file = os.path.expanduser("~/.tmux.conf")
        if self.conf_file == default_conf_file:
            self.check_replace_default_config()
        elif os.path.exists(self.conf_file):
            self.check_replace_custom_config()

    def check_replace_default_config(self) -> None:
        if os.path.isfile(self.conf_file) or os.path.islink(self.conf_file):
            confirmation = input(
                "Do you wish to replace the default config file (y/n)?"
            )
        else:
            confirmation = input("Do you wish to create a default config file (y/n)?")
        if confirmation not in ("y", "Y"):
            print("Terminating...")
            sys.exit(1)

    def check_replace_custom_config(self) -> None:
        confirmation = input(f"Do you wish to replace {self.conf_file} (y/n)?")
        if confirmation not in ("y", "Y"):
            print("Terminating...")
            sys.exit(1)

    #
    #  Selecting tmux bin
    #
    def find_tmux_bin(self, cmd: str = "", requested_vers: str = "") -> None:
        """if tmux should use a specific version, when generating config
        set requested_vers parameter, if empy the actual version is used
        if found, will set self.tmux_bin and more
        if not Raises exception
        """
        if not cmd:
            #  Use the first that gives something
            cmd = self.find_cmd_1() or self.full_path_cmd() or self.find_cmd_3()

        if not cmd:
            raise TmuxConfNotTmuxCommand("No tmux command found")
        self.expand_cmd_path(cmd, requested_vers)

        # ==== 1 =

        if not self.tmux_bin:
            #  tmux not found abort
            print("ERROR could not find tmux binary, aborting")
            sys.exit(1)
        return

    def find_cmd_1(self) -> str:
        """First check is to see if the tmux used to generate the
        previous config can be extracted.
        """
        cmd = ""
        if os.path.isfile(self.conf_file):
            s = f'grep "TMUX_BIN=" {self.conf_file} | cut -d= -f 2 | sed s/\\"//g'
            c = run_shell(s).strip()
            if not c:
                return ""
            c2 = shutil.which(c)  # try to expand full path if not given
            if not c2:
                return ""
            cmd = c2
            print(f"found {cmd} in conf file")
        return cmd

    def find_cmd_3(self) -> str:
        """Finally try some likely locations"""
        cmd = ""
        for c in (
            "/usr/local/bin/tmux",
            "/home/linuxbrew/.linuxbrew/bin/tmux",
            "~/.asdf/shims/tmux",
        ):
            c2 = expanduser_plus(c)
            if is_executable(c2):
                cmd = c2
                print(f"found {cmd} manually")
                break
        return cmd

    def use_tmux_bin(self, tmux_bin: str = "tmux", requested_vers: str = "") -> bool:
        """Returns True if this was a valid tmux bin
        If this was the case self.tmux_bin & self.vers will have been set
        """
        parts = run_shell(f"{tmux_bin} -V").split(" ")
        if len(parts) != 2 or parts[0] not in ("tmux", "tmate"):
            raise TmuxConfNotTmuxCommand(f"{tmux_bin} Doesnt seem to be a tmux binary")

        vers = VersionCheck(actual_vers=parts[1], requested_vers=requested_vers)
        # except TmuxConfInvalidTmuxVersion:
        #     print("{parts[[1]} Doesnt seem to be a valid tmux version")
        #     return False
        self.tmux_bin = tmux_bin
        # pylint: disable=W0201
        self.vers = vers
        return True

    #
    #  Various attempt at finding the tmux bin
    #
    def expand_cmd_path(self, cmd: str, requested_vers: str = "") -> None:
        """If this is an asdf tmux, we need to translate path from shim
        to actual bin, in order to ptoperly detect version, but make sure
        not to expand an asdf path that has already been processed!

        Any process starting in the same dir as tmux was started will
        inherrit potential local asdf tmux selection, but processes
        starting in other dirs will use asdf default version.
        By expanding the initial shim into a full path,
        the correct tmux will be used in all cases.
        """
        if cmd:
            #  First use what in the asdf case might be a shim, in order
            #  to get full path
            #
            if not self.use_tmux_bin(cmd, requested_vers):
                self.tmux_bin = ""
                return
        # get full path notation for tmux
        cmd = self.full_path_cmd(self.tmux_bin)
        if cmd.find(".asdf") > -1 and cmd.find("/installs/") < 0:
            if not self.use_tmux_bin(cmd):
                print(f"ERROR: asdf tmux does not seem to be valid: {cmd}")
                sys.exit(1)
            #
            #  Convert asdf shim into absolute path
            #
            cmd_asdf = (
                f'{cmd.split("shims/")[0]}installs/tmux/'
                f'{self.vers.get().split("-", maxsplit=1)[0]}/bin/tmux'
            )
            self.use_tmux_bin(cmd_asdf)

    def full_path_cmd(self, cmd: str = "tmux") -> str:
        c = run_shell(f"command -v {cmd}")
        if c and c.lower().find("not found") < 0:
            cmd = c
            print(f"found {cmd} in PATH")
        return cmd

    def remove_conf_file(self) -> None:
        if os.path.exists(self.conf_file):
            #  Ensure we start with an empty file
            os.remove(self.conf_file)

    def is_tmate(self) -> bool:
        return self.tmux_bin.find("tmate") > -1



#  -*- mode: python; mode: fold -*-
#
#  Copyright (c) 2022-2024: Jacob.Lundqvist@gmail.com
#  License: MIT
#
#  Part of https://github.com/jaclu/tmux-conf
#
#  See constants.py for version info
#
#  See the README.md in the repository for more info
#

"""Class that handles embedded scripts"""

import os
import pathlib
import stat
import sys

from .constants import XDG_CONFIG_HOME
from .utils import run_shell, tilde_home_dir
from .vers_check import VersionCheck


class EmbeddedScripts:
    """Handles scripts, either embedded or stored in scripts/ as external"""

    def __init__(
        self, conf_file: str, vers_class: VersionCheck, use_embedded_scripts: bool
    ) -> None:
        #  Ensure conf file is using ~ or full path if ~ not applicable
        conf_file = tilde_home_dir(conf_file)
        if conf_file[0] not in ("~", "/"):
            conf_file = tilde_home_dir(os.path.join(os.getcwd(), conf_file))
        self._conf_file = conf_file
        self._vers = vers_class
        self._use_embedded_scripts = use_embedded_scripts
        self._scripts: list[str] = []
        self.defined_scripts: list[str] = []
        self._bash_scripts: list[str] = []
        self._bash_shell = ""  # Will only be set if needed

    def create(
        self,
        scr_name: str,
        script: list[str],
        use_bash: bool = False,
        built_in: bool = False,
    ) -> None:
        """Creates a script, supplied as a list of lines
        script lines can be regular lines as string, or multi-line strings

        built_in is set to True for scripts generated in this pip
        then it is checked if such a script has already been created by
        code using this, and if the scr_name has been used, this instance
        of it is skipped

        Where the script should be run, just call run_it() with
        the script name, and code will be generated to call it the right way.

        calling self.run_it('activate_tpm')  will (depending on config file)
        insert code like:
          run "cut -c3- /home/jaclu/.tmux.conf | sh -s activate_tpm"
        or:
          run "/home/jaclu/.tmux/scripts/activate_tpm.sh"

        depending on self._use_embedded_scripts
        """
        if built_in is False:
            self.defined_scripts.append(scr_name)
        else:
            #  Allow users to override default scripts
            if scr_name in self.defined_scripts:
                return
        if self._use_embedded_scripts:
            if use_bash:
                self._bash_scripts.append(scr_name)
            self._scripts += script
            self._scripts.append("")  # separator between scripts
        else:
            script_dir = self.get_dir()
            os.makedirs(script_dir, exist_ok=True)
            if use_bash:
                shebang = "#!/usr/bin/env bash"
            else:
                shebang = "#!/bin/sh"

            script.insert(0, f"{shebang}\n")
            script.append(f'{scr_name} "$@"')  # trigger it to run without a param

            fname = f"{script_dir}/{scr_name}.sh"
            with open(fname, "w", encoding="utf-8") as f:
                for line in script:
                    f.write(f"{line}\n")

            #  Make it run able
            f2 = pathlib.Path(fname)
            f2.chmod(f2.stat().st_mode | stat.S_IEXEC | stat.S_IXGRP | stat.S_IXOTH)

    def run_it(self, scr_name: str, in_bg: bool = False) -> str:
        """Generate the code to run an embedded/external script"""
        cmd = "run-shell "
        if self._vers.is_ok(1.8) and in_bg:
            cmd += "-b "
        cmd += '"'
        if self._use_embedded_scripts:
            cmd += f"cut -c3- {self._conf_file} | "
            if scr_name in self._bash_scripts:
                if not self._bash_shell:
                    bash_shell = run_shell("command -v bash")
                    if not bash_shell:
                        sys.exit("Failed to find bash!")
                    self._bash_shell = bash_shell
                cmd += self._bash_shell
            else:
                cmd += "sh"
            cmd += f" -s {scr_name}"
        else:
            cmd += f"{self.get_dir()}/{scr_name}.sh"
        cmd += '"'
        return cmd

    def call_script(self, scr_name: str) -> str:
        """call a script"""
        cmd = scr_name
        if not self._use_embedded_scripts:
            cmd = f"{self.get_dir()}/{scr_name}.sh"
        return cmd

    def content(self):
        """Generates the code for embedded scripts to be written to
        the conf file"""
        if not (self._use_embedded_scripts and self._scripts):
            return []

        output = [
            """
        #======================================================
        #
        # EMBEDDED-SCRIPTS-STARTING-POINT
        #"""
        ]
        #
        #  Embedded scripts starts at column 3, insert extra indention on
        #  each line
        #
        for script_line in self._scripts:
            mutlilines = script_line.split("\n")
            if len(mutlilines) > 1:
                #
                #  Was multi-line string, remove first and last lines if empty
                #
                if not mutlilines[0]:
                    mutlilines.pop(0)
                if not mutlilines[-1]:
                    mutlilines.pop(-1)
                for line in mutlilines:
                    output.append(f"# {line}")
            else:
                output.append(f"# {script_line}")
        output.append('# "$@" #  This triggers the embedded script')
        return output

    def get_dir(self) -> str:
        """Retrieves the location where scripts should be saved"""
        if self._use_embedded_scripts:
            raise SyntaxError("get_dir() called when use_embedded_scripts is True")

        if tilde_home_dir(self._conf_file) == "~/.tmux.conf":
            scripts_dir = os.path.expanduser("~/.tmux/scripts")
        else:
            conf_file = os.path.expanduser(self._conf_file)

            xdg_home = os.environ.get(XDG_CONFIG_HOME)
            if xdg_home:
                conf_base = xdg_home
            else:
                conf_base = os.path.dirname(os.path.dirname(conf_file))

            scripts_dir = os.path.expanduser(os.path.join(conf_base, "tmux", "scripts"))
        return scripts_dir



#  -*- mode: python; mode: fold -*-
#
#  Copyright (c) 2022-2024: Jacob.Lundqvist@gmail.com
#  License: MIT
#
#  Part of https://github.com/jaclu/tmux-conf
#
#  See constants.py for version info
#
#  Class that handles version checking
#
#  See the README.md in the repository for more info
#
"""compares tmux versions"""


class VersionCheck:
    """compares tmux versions"""

    def __init__(self, actual_vers, requested_vers=None):
        self._vers_actual = self.normalize_vers(actual_vers)
        if requested_vers:
            self._vers = self.normalize_vers(requested_vers)
        else:
            self._vers = self._vers_actual
        #
        #  For performance, the version parts are pre-calculated
        #

        #
        #  tmux derivative tmate uses versions like 2.4.0
        #  for compatibility checks only the first two are needed
        #  This syntax handles both normal tmux and tmate
        #
        parts = self._vers.split(".")
        v_maj = parts[0]
        v_min = parts[1]
        try:
            self.v_maj = int(v_maj)
        except ValueError as exc:
            print(f"Error: v_maj was not int: {self._vers}")
            raise ValueError from exc
        self.v_min, self.v_suffix = self.get_sub_vers(v_min)

    def get(self):
        """The version used for generating the config"""
        return self._vers

    def get_actual(self):
        """The version of the tmux bin"""
        return self._vers_actual

    def is_ok(self, vers) -> bool:
        """Checks version vs current tmux environment
        Param is forgiving, can be int, float or string.
        When given as int .0 is appended
        In many cases a float is sufficient, like 2.8, 3.0 etc
        Some versions have character suffixes like 3.3a, then a string
        param is needed. Internally version refs are allways treated as
        strings.
        """
        a, b = self.normalize_vers(vers).split(".")

        try:
            vers_maj = int(a)
        except ValueError as exc:
            print(f"ERROR: vers_ok({vers}) - maj part not int!")
            raise ValueError from exc

        vers_min, suffix = self.get_sub_vers(b)

        if vers_maj > self.v_maj:
            return False
        if vers_maj < self.v_maj:
            return True

        r = True
        if vers_min > self.v_min:
            r = False
        elif vers_min == self.v_min and suffix > self.v_suffix:
            r = False
        return r

    def get_sub_vers(self, v2: str):
        """get sub version"""

        int_part = ""
        for c in v2:
            try:
                int(c)
            except ValueError:
                break
            int_part += c
        if not int_part:
            raise ValueError("sub_vers had no int part")
        i = int(int_part)
        s = v2.split(int_part)[1]
        return i, s

    def normalize_vers(self, vers) -> str:
        """Normalizes vers into a string"""
        # param fixes

        if vers == "next-3.4":
            #
            #  Alpine tmux-3.3a_git20230428 reports version as next-3.4
            #  Change this into 3.3a-git20230428, so that notation follows
            #  the normal logic, and can be used in vers checks
            #
            vers = "3.3a-git20230428"
        elif vers == "3.1-rc":
            #
            #  asdf tmux 3.1 reports as 3.1-rc, for simplicity just
            #  brand it as 3.1
            #
            vers = "3.1"

        if isinstance(vers, str) and vers.find(".") < 0:
            try:
                vers = int(vers)
            except ValueError as err:
                print(f"ERROR: vers_check normalize_vers({vers}) bad param")
                raise ValueError from err
        if isinstance(vers, int):
            vers = f"{vers}.0"
        elif isinstance(vers, float):
            vers = f"{vers}"
        #  correct , -> .
        vers = vers.replace(",", ".")
        #
        #  Only keep first two items
        #
        parts = vers.split(".")
        # if len(parts) > 2:
        #     raise ValueError(f"ERROR: normalize_vers({vers}) > 2 parts")
        if len(parts) < 2:
            raise ValueError(f"ERROR: normalize_vers({vers}) < 2 parts")
        return ".".join(parts[:2])



#  -*- mode: python; mode: fold -*-
#
#  Copyright (c) 2022-2024: Jacob.Lundqvist@gmail.com
#  License: MIT
#
#  Part of https://github.com/jaclu/tmux-conf
#
#  See the README.md in the repository for more info
#
"""packet wide constants"""

__version__ = "0.17.3"

XDG_CONFIG_HOME = "XDG_CONFIG_HOME"



import pytest
from src.tmux_conf.vers_check import VersionCheck


def not_test_vc_maj_not_int():
    with pytest.raises(ValueError):
        VersionCheck("a.b")


def not_test_vc_is_ok_maj_not_int():
    vc = VersionCheck(3.2)
    with pytest.raises(ValueError):
        vc.is_ok("a")


def not_test_vc_bad_min():
    with pytest.raises(ValueError):
        VersionCheck("3.b")


def not_test_vc_same():
    vc = VersionCheck(3.2)
    assert vc.is_ok(3.2) is True  # noqa


def not_test_vc_same_suffix():
    vc = VersionCheck("3.2a")
    assert vc.is_ok("3.2a") is True  # noqa


def not_test_vc_maj_high():
    vc = VersionCheck(3.1)
    assert vc.is_ok(4.0) is not True  # noqa


def not_test_vc_maj_high_int_str():
    vc = VersionCheck(3.0)
    assert vc.is_ok("4") is not True


def not_test_vc_maj_low():
    vc = VersionCheck(3.1)
    assert vc.is_ok(2.0) is True  # noqa


def not_test_vc_min_high():
    vc = VersionCheck(3.2)
    assert vc.is_ok(3.3) is not True  # noqa


def not_test_vc_min_low():
    vc = VersionCheck(3.2)
    assert vc.is_ok(3.0) is True  # noqa


def not_test_vc_suffix_high():
    vc = VersionCheck("3.1a")
    assert vc.is_ok("3.1b") is not True  # noqa


def not_test_vc_suffix_low():
    vc = VersionCheck("3.1b")
    assert vc.is_ok("3.1a") is True  # noqa


def not_test_vc_req_actual():
    vc = VersionCheck(3.0, 2.4)
    assert vc.get_actual() == "3.0"


def not_test_vc_req_get():
    vc = VersionCheck(3.0, 2.4)
    assert vc.get() == "2.4"


def not_test_vc_req_same():
    vc = VersionCheck(2.0, 2.4)
    assert vc.is_ok(2.4) is True


def not_test_vc_req_same_suffix():
    vc = VersionCheck(2.0, "2.4a")
    assert vc.is_ok("2.4a") is True


def not_test_vc_req_high():
    vc = VersionCheck(2.0, 2.4)
    assert vc.is_ok(2.5) is not True


def not_test_vc_req_high_suffix():
    vc = VersionCheck(2.0, 2.4)
    assert vc.is_ok("2.4a") is not True


def not_test_vc_req_inbetween():
    vc = VersionCheck(2.0, 2.4)
    assert vc.is_ok(2.2) is True


def not_test_vc_req_inbetween_suffix():
    vc = VersionCheck(2.0, 2.4)
    assert vc.is_ok("2.2a") is True


def not_test_vc_req_low():
    vc = VersionCheck(3.0, 2.4)
    assert vc.is_ok(2.3) is True


def not_test_vc_req_low_suffix():
    vc = VersionCheck(3.0, 2.4)
    assert vc.is_ok("2.3a") is True


def not_test_vc_req_low_both():
    vc = VersionCheck(3.0, 2.4)
    assert vc.is_ok(2.2) is True


def not_test_vc_req_low_both_suffix():
    vc = VersionCheck(3.0, 2.4)
    assert vc.is_ok("2.2a") is True


def not_test_vc_missing_subvers():
    vc = VersionCheck(3)
    assert vc.is_ok(3.0) is True


def not_test_vc_missing_subvers_str():
    vc = VersionCheck("3")
    assert vc.is_ok(3.0) is True


def not_test_vc_normalize_vers():
    vc = VersionCheck(3.0)
    assert vc.normalize_vers(2) == "2.0"


def not_test_vc_vers_bad_init():
    with pytest.raises(ValueError):
        VersionCheck("qwerty")



#  -*- mode: python; mode: fold -*-
#
#  Copyright (c) 2022-2024: Jacob.Lundqvist@gmail.com
#  License: MIT
#
#  Part of https://github.com/jaclu/tmux-conf
#
#  See constants.py for version info
#
#  See the README.md in the repository for more info
#
#  All paths stored internally are done so with ~ expanded
#

"""Common stuff"""

import argparse
import os
import subprocess  # nosec B404

from .constants import __version__


def btick_unescaped(line: str) -> bool:
    """Ensure any back-ticks on the line are escaped,
    otherwise running embedded scripts will fail.
    """
    pos = line.find("`")
    if not line or pos < 0:
        return False  # None present
    if line[pos - 1] != "\\":
        return True  # Was nor escaped
    #
    # recurse to check rest of l§ine
    #
    resume_pos = pos + 1
    return btick_unescaped(line[resume_pos:])


def expanduser_plus(path: str) -> str:
    """Handles both ~ and $HOME"""
    return os.path.expanduser(os.path.expandvars(path))


def is_executable(cmd: str) -> bool:
    """check if cmd is executable"""
    fpath = os.path.expanduser(cmd)
    return os.path.isfile(fpath) and os.access(fpath, os.X_OK)


def parse_cmdline(args):
    """Parse command line options"""

    parser = argparse.ArgumentParser(
        description="This tmux conf compiler generates configs for tmux "
        + "versions 1.5 and up."
    )

    #  Prints tmux-conf version info then exits
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s using tmux-conf: {__version__}",
    )

    parser.add_argument(
        "-c",
        "--clear_plugins",
        action="store_true",
        help="Removes tpm and current plugins. "
        + "They will be reinstalled on next launch",
    )

    parser.add_argument(
        "-r",
        "--replace",
        action="store_true",
        help="Do not ask for confirmation before replacing a config file",
    )

    parser.add_argument(
        "-p",
        "--plugins_display",
        choices=[1, 2, 3],
        default=0,
        type=int,
        help="1=Used plugins 2=also ignored plugins 3=show config",
    )

    parser.add_argument(
        "-t", "--tmux_bin", help="Force usage of this tmux bin", default=""
    )

    parser.add_argument(
        "-V",
        "--forced-version",
        help="Hard-code tmux version to be assumed during process",
    )

    parser.add_argument(
        "conf_file",
        nargs="?",
        default="~/.tmux.conf",
        help="Config file to be created, Defaults to ~/.tmux.conf",
    )
    return parser.parse_args(args)


def run_shell(cmd: str) -> str:
    """Run a command in a shell"""
    # pylint: disable=subprocess-run-check
    result = subprocess.run(
        cmd, capture_output=True, text=True, shell=True  # nosec: B602
    )
    return result.stdout.strip()


def tilde_home_dir(path: str) -> str:
    """Converts full path to ~ notation
    Internally all path related variables are stored expanded for python
    readability. I try to allways use this when writing paths to the
    config file, to keep it generic"""
    path = os.path.expanduser(path)
    home_dir = os.path.expanduser("~")
    # pylint: disable=compare-to-zero
    if path.find(home_dir) == 0:
        r = path.replace(home_dir, "~")
    else:
        r = path
    return r


def verify_conf_file_usable(conf_file: str) -> str:
    """Verify conf_file can be written to"""
    if not conf_file:
        raise SyntaxError("ERROR: Empty conf_file param!")

    # FILES
    # ~/.tmux.conf
    # $XDG_CONFIG_HOME/tmux/tmux.conf
    # ~/.config/tmux/tmux.conf     Default tmux configuration file.
    # /usr/local/etc/tmux.conf     System-wide configuration file.

    #  Ensure it can be parsed by Python
    conf_file = expanduser_plus(conf_file)

    #  Ensure conf_file is referred to with full path
    if not os.path.dirname(conf_file):
        conf_file = os.path.join(os.getcwd(), conf_file)

    conf_location = os.path.dirname(conf_file)
    try:
        os.makedirs(conf_location, exist_ok=True)
    except PermissionError as error:
        print("ERROR: Could not create directory for config file!")
        raise PermissionError from error
    except OSError as error:
        print(f"ERROR: read only file system reported for {conf_file}")
        raise OSError from error

    #  Ensure it can be written to
    try:
        with open(conf_file, "a", encoding="utf-8") as _:
            pass
    except IOError as error:
        print(f"ERROR: Could not write to config file: {conf_file}")
        raise IOError from error

    #  If it didn't exist, we just created an empty file, get rid of it
    if os.stat(conf_file).st_size < 1:
        os.remove(conf_file)

    #  Return full path notation for config file
    return conf_file



#  -*- mode: python; mode: fold -*-
#
#  Copyright (c) 2022-2024: Jacob.Lundqvist@gmail.com
#  License: MIT
#
#  Part of https://github.com/jaclu/tmux-conf
#
#  See the README.md in the repository for more info
#
"""tmux-conf exceptions"""


class TmuxConfNotTmuxCommand(Exception):
    """Command does not seem to be tmux"""

    def __init__(self, message="Invalid tmux command"):
        self.message = message
        super().__init__(self.message)


class TmuxConfInvalidTmuxVersion(Exception):
    """Version is not valid tmux version"""

    def __init__(self, message="Invalid tmux version string"):
        self.message = message
        super().__init__(self.message)
